@@ -24,6 +24,7 @@
 #include <QSGSimpleTextureNode>
 #include <QSGTexture>
 #include <QScreen>
+#include <cstdlib>
 
 Q_GLOBAL_STATIC(ImageTexturesCache, s_iconImageCache)
 
@@ -264,6 +265,11 @@ QSGNode *Icon::updatePaintNode(QSGNode *node, QQuickItem::UpdatePaintNodeData *
             }
         }
 
+        // Adjust the final node on the pixel grid
+        QPointF globalPixelPos = mapToScene(nodeRect.topLeft()) * m_devicePixelRatio;
+        QPointF posAdjust = QPointF(globalPixelPos.x() - std::round(globalPixelPos.x()), globalPixelPos.y() - std::round(globalPixelPos.y()));
+        nodeRect.moveTopLeft(nodeRect.topLeft() - posAdjust);
+
         for (int i = 0; i < node->childCount(); ++i) {
             auto mNode = static_cast<ManagedTextureNode *>(node->childAtIndex(i)->firstChild());
             mNode->setRect(nodeRect);
