@@ -12,6 +12,7 @@
 #include <QGuiApplication>
 #include <QQmlComponent>
 #include <QQmlEngine>
+#include <QQuickStyle>
 #include <QStyleHints>
 
 #include <chrono>
@@ -242,11 +243,15 @@ int Units::maximumInteger() const
 Units *Units::create(QQmlEngine *qmlEngine, [[maybe_unused]] QJSEngine *jsEngine)
 {
 #ifndef KIRIGAMI_BUILD_TYPE_STATIC
-    auto plugin = PlatformPluginFactory::findPlugin();
+    const QString pluginName = qmlEngine->property("_kirigamiTheme").toString();
+
+    auto plugin = PlatformPluginFactory::findPlugin(pluginName);
+    if (!plugin && !pluginName.isEmpty()) {
+        plugin = PlatformPluginFactory::findPlugin();
+    }
+
     if (plugin) {
         return plugin->createUnits(qmlEngine);
-    } else {
-        qWarning(KirigamiPlatform) << "Failed to find a Kirigami platform plugin";
     }
 #endif
     // Fall back to the default units implementation
